use super::file::FileWriter;
use crate::error::ArchiveError;
use crate::Handle;
use crate::{WriteFilter, WriteFormat};
use libarchive3_sys::ffi;
use std::ffi::CString;
use std::path::Path;

pub struct Builder {
    handle: *mut ffi::Struct_archive,
    consumed: bool,
}

unsafe impl Send for Builder {}

impl Builder {
    pub fn new() -> Self {
        Builder::default()
    }

    pub fn add_filter(&mut self, filter: WriteFilter) -> crate::Result<()> {
        let result = match filter {
            WriteFilter::B64Encode => unsafe {
                ffi::archive_write_add_filter_b64encode(self.handle)
            },
            WriteFilter::Bzip2 => unsafe { ffi::archive_write_add_filter_bzip2(self.handle) },
            WriteFilter::Compress => unsafe { ffi::archive_write_add_filter_compress(self.handle) },
            WriteFilter::Grzip => unsafe { ffi::archive_write_add_filter_grzip(self.handle) },
            WriteFilter::Gzip => unsafe { ffi::archive_write_add_filter_gzip(self.handle) },
            WriteFilter::Lrzip => unsafe { ffi::archive_write_add_filter_lrzip(self.handle) },
            WriteFilter::Lzip => unsafe { ffi::archive_write_add_filter_lzip(self.handle) },
            WriteFilter::Lzma => unsafe { ffi::archive_write_add_filter_lzma(self.handle) },
            WriteFilter::Lzop => unsafe { ffi::archive_write_add_filter_lzop(self.handle) },
            WriteFilter::None => unsafe { ffi::archive_write_add_filter_none(self.handle) },
            WriteFilter::Program(prog) => {
                let c_prog = CString::new(prog).unwrap();
                unsafe { ffi::archive_write_add_filter_program(self.handle, c_prog.as_ptr()) }
            }
            WriteFilter::UuEncode => unsafe { ffi::archive_write_add_filter_uuencode(self.handle) },
            WriteFilter::Xz => unsafe { ffi::archive_write_add_filter_xz(self.handle) },
        };
        match result {
            ffi::ARCHIVE_OK => Ok(()),
            _ => Result::from(self as &dyn Handle),
        }
    }

    pub fn set_format(&self, format: WriteFormat) -> crate::Result<()> {
        let result = match format {
            WriteFormat::SevenZip => unsafe { ffi::archive_write_set_format_7zip(self.handle) },
            WriteFormat::ArBsd => unsafe { ffi::archive_write_set_format_ar_bsd(self.handle) },
            WriteFormat::ArSvr4 => unsafe { ffi::archive_write_set_format_ar_svr4(self.handle) },
            WriteFormat::Cpio => unsafe { ffi::archive_write_set_format_cpio(self.handle) },
            WriteFormat::CpioNewc => unsafe {
                ffi::archive_write_set_format_cpio_newc(self.handle)
            },
            WriteFormat::Gnutar => unsafe { ffi::archive_write_set_format_gnutar(self.handle) },
            WriteFormat::Iso9660 => unsafe { ffi::archive_write_set_format_iso9660(self.handle) },
            WriteFormat::Mtree => unsafe { ffi::archive_write_set_format_mtree(self.handle) },
            WriteFormat::MtreeClassic => unsafe {
                ffi::archive_write_set_format_mtree_classic(self.handle)
            },
            WriteFormat::Pax => unsafe { ffi::archive_write_set_format_pax(self.handle) },
            WriteFormat::PaxRestricted => unsafe {
                ffi::archive_write_set_format_pax_restricted(self.handle)
            },
            WriteFormat::Shar => unsafe { ffi::archive_write_set_format_shar(self.handle) },
            WriteFormat::SharDump => unsafe {
                ffi::archive_write_set_format_shar_dump(self.handle)
            },
            WriteFormat::Ustar => unsafe { ffi::archive_write_set_format_ustar(self.handle) },
            WriteFormat::V7tar => unsafe { ffi::archive_write_set_format_v7tar(self.handle) },
            WriteFormat::Xar => unsafe { ffi::archive_write_set_format_xar(self.handle) },
            WriteFormat::Zip => unsafe { ffi::archive_write_set_format_zip(self.handle) },
        };
        match result {
            ffi::ARCHIVE_OK => Ok(()),
            _ => Result::from(self as &dyn Handle),
        }
    }

    pub fn open_file<T: AsRef<Path>>(mut self, file: T) -> crate::Result<FileWriter> {
        if self.consumed {
            return Err(ArchiveError::Consumed);
        }
        let c_file = CString::new(file.as_ref().to_string_lossy().as_bytes()).unwrap();
        let res = unsafe { ffi::archive_write_open_filename(self.handle, c_file.as_ptr()) };
        match res {
            ffi::ARCHIVE_OK => {
                self.consumed = true;
                Ok(FileWriter::new(self.handle))
            }
            _ => Err(ArchiveError::from(&self as &dyn Handle)),
        }
    }
}

impl Default for Builder {
    fn default() -> Self {
        unsafe {
            let handle = ffi::archive_write_new();
            if handle.is_null() {
                panic!("Allocation error");
            }
            Builder {
                handle,
                consumed: false,
            }
        }
    }
}

impl Handle for Builder {
    unsafe fn handle(&self) -> *const ffi::Struct_archive {
        self.handle as *const _
    }

    unsafe fn handle_mut(&mut self) -> *mut ffi::Struct_archive {
        self.handle
    }
}

impl Drop for Builder {
    fn drop(&mut self) {
        if !self.consumed {
            unsafe {
                ffi::archive_write_free(self.handle);
            }
        }
    }
}
