use super::WriteEntry;
use crate::error::ArchiveError;
use crate::Entry;
use crate::Handle;
use core::ffi::c_void;
use libarchive3_sys::ffi;
use std::fs;
use std::io;
use std::io::Read;
use std::io::Write;
use std::path::Path;

pub struct FileWriter {
    handle: *mut ffi::Struct_archive,
    closed: bool,
}

unsafe impl Send for FileWriter {}

impl Handle for FileWriter {
    unsafe fn handle(&self) -> *const ffi::Struct_archive {
        self.handle as *const _
    }

    unsafe fn handle_mut(&mut self) -> *mut ffi::Struct_archive {
        self.handle
    }
}

impl FileWriter {
    pub fn new(handle: *mut ffi::Struct_archive) -> Self {
        FileWriter {
            handle,
            closed: false,
        }
    }

    /// Append the given entry to the archive. After successfully calling this function, writing to
    /// the archive now writes to this entry.
    pub fn append_entry(&mut self, entry: &mut WriteEntry) -> crate::Result<()> {
        unsafe {
            match ffi::archive_write_header(self.handle_mut(), entry.entry_mut()) {
                ffi::ARCHIVE_OK => Ok(()),
                _ => Err(ArchiveError::from(self as &dyn Handle).into()),
            }
        }
    }

    pub fn append_data<R: Read>(&mut self, entry: &mut WriteEntry, r: &mut R) -> crate::Result<()> {
        unsafe {
            match ffi::archive_write_header(self.handle_mut(), entry.entry_mut()) {
                ffi::ARCHIVE_OK => (),
                _ => return Err(ArchiveError::from(self as &dyn Handle).into()),
            }
        }

        let mut buf = [0; 4096];

        loop {
            match r.read(&mut buf) {
                Ok(0) => return Ok(()),
                // Write entire buffer
                Ok(buf_len) => {
                    let mut written: usize = 0;

                    while written < buf_len {
                        let res = unsafe {
                            ffi::archive_write_data(
                                self.handle_mut(),
                                &buf[written] as *const u8 as *const _,
                                buf_len - written,
                            )
                        } as isize;

                        // Negative values signal errors
                        if res < 0 {
                            return Err(ArchiveError::from(self as &dyn Handle).into());
                        }

                        written += usize::try_from(res).unwrap();
                    }
                }
                Err(err) => match err.kind() {
                    io::ErrorKind::Interrupted => (),
                    _ => return Err(err.into()),
                },
            };
        }
    }

    pub fn append_path<P: AsRef<Path>>(
        &mut self,
        entry: &mut WriteEntry,
        path: P,
    ) -> crate::Result<()> {
        let mut f = fs::File::open(path)?;
        self.append_data(entry, &mut f)
    }

    pub fn close(&mut self) -> crate::Result<()> {
        unsafe {
            match ffi::archive_write_close(self.handle_mut()) {
                ffi::ARCHIVE_OK => {
                    self.closed = true;

                    Ok(())
                }
                _ => Err(ArchiveError::from(self as &dyn Handle)),
            }
        }
    }
}

impl Drop for FileWriter {
    fn drop(&mut self) {
        if !self.closed {
            let _ = self.close();
        }

        unsafe {
            ffi::archive_write_free(self.handle_mut());
        }
    }
}

impl Write for FileWriter {
    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
        let res = unsafe {
            ffi::archive_write_data(self.handle_mut(), buf.as_ptr() as *const c_void, buf.len())
        } as isize;

        if res < 0 {
            Err(ArchiveError::from(self as &dyn Handle).into())
        } else {
            // Unwrap is safe as we check if the value is negative in the if statement
            Ok(res.try_into().unwrap())
        }
    }

    fn flush(&mut self) -> io::Result<()> {
        // Libarchive doesn't seem to provide a flush mechanism
        Ok(())
    }
}
