use crate::{repo::MetaRepoMgr, Config, Global};

use axum::{extract::FromRef, Router};
use clap::Parser;
use sea_orm_migration::MigratorTrait;
use std::{io, path::PathBuf, sync::Arc};
use tokio::sync::RwLock;
use tower_http::trace::TraceLayer;
use tracing::debug;
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
pub struct Cli {
    /// Directory where repository metadata & SQLite database is stored
    #[arg(env = "RIETER_DATA_DIR")]
    pub data_dir: PathBuf,
    /// API key to authenticate private routes with
    #[arg(env = "RIETER_API_KEY")]
    pub api_key: String,

    /// Database connection URL; either sqlite:// or postgres://. Defaults to rieter.sqlite in the
    /// data directory
    #[arg(short, long, env = "RIETER_DATABASE_URL")]
    pub database_url: Option<String>,
    /// Port the server will listen on
    #[arg(
        short,
        long,
        value_name = "PORT",
        default_value_t = 8000,
        env = "RIETER_PORT"
    )]
    pub port: u16,
    /// Log levels for the tracing
    #[arg(
        long,
        value_name = "LOG_LEVEL",
        default_value = "tower_http=debug,rieterd=debug,sea_orm=debug",
        env = "RIETER_LOG"
    )]
    pub log: String,
}

impl FromRef<Global> for Arc<RwLock<MetaRepoMgr>> {
    fn from_ref(global: &Global) -> Self {
        Arc::clone(&global.repo_manager)
    }
}

impl Cli {
    pub fn init_tracing(&self) {
        tracing_subscriber::registry()
            .with(tracing_subscriber::EnvFilter::new(self.log.clone()))
            .with(tracing_subscriber::fmt::layer())
            .init();
    }

    pub async fn run(&self) -> crate::Result<()> {
        self.init_tracing();

        let db_url = if let Some(url) = &self.database_url {
            url.clone()
        } else {
            format!(
                "sqlite://{}?mode=rwc",
                self.data_dir.join("rieter.sqlite").to_string_lossy()
            )
        };

        debug!("Connecting to database with URL {}", db_url);

        let mut options = sea_orm::ConnectOptions::new(db_url);
        options.max_connections(16);

        let db = sea_orm::Database::connect(options).await?;
        crate::db::Migrator::up(&db, None).await?;

        debug!("Successfully applied migrations");

        let config = Config {
            data_dir: self.data_dir.clone(),
        };
        let repo_manager = MetaRepoMgr::new(&self.data_dir.join("repos"));

        let pkg_filename_re = regex::Regex::new(r"^([a-z0-9@_+][a-z0-9@_+-.]*)-((?:[0-9]+:)?[a-z0-9._]+-[0-9.]+)-([a-z0-9_]+)\.pkg\.tar\.([a-z0-9]+)$").unwrap();

        let global = Global {
            config,
            repo_manager: Arc::new(RwLock::new(repo_manager)),
            db,
            pkg_filename_re,
        };

        // build our application with a single route
        let app = Router::new()
            .nest("/api", crate::api::router())
            .merge(crate::repo::router(&self.api_key))
            .with_state(global)
            .layer(TraceLayer::new_for_http());

        let domain: String = format!("0.0.0.0:{}", self.port).parse().unwrap();
        let listener = tokio::net::TcpListener::bind(domain).await?;
        // run it with hyper on localhost:3000
        Ok(axum::serve(listener, app.into_make_service())
            .await
            .map_err(|err| io::Error::new(io::ErrorKind::Other, err))?)
    }
}
