mod archive;
mod manager;
mod manager_new;
pub mod package;

pub use manager_new::MetaRepoMgr;

use axum::{
    body::Body,
    extract::{Path, State},
    http::{Request, StatusCode},
    response::IntoResponse,
    routing::{delete, post},
    Router,
};
use futures::TryStreamExt;
use tokio_util::io::StreamReader;
use tower::util::ServiceExt;
use tower_http::{services::ServeFile, validate_request::ValidateRequestHeaderLayer};

pub fn router(api_key: &str) -> Router<crate::Global> {
    Router::new()
        .route(
            "/:repo",
            post(post_package_archive)
                .delete(delete_repo)
                .route_layer(ValidateRequestHeaderLayer::bearer(api_key)),
        )
        .route(
            "/:repo/:arch",
            delete(delete_arch_repo).route_layer(ValidateRequestHeaderLayer::bearer(api_key)),
        )
        // Routes added after the layer do not get that layer applied, so the GET requests will not
        // be authorized
        .route(
            "/:repo/:arch/:filename",
            delete(delete_package)
                .route_layer(ValidateRequestHeaderLayer::bearer(api_key))
                .get(get_file),
        )
}

/// Serve the package archive files and database archives. If files are requested for an
/// architecture that does not have any explicit packages, a repository containing only "any" files
/// is returned.
async fn get_file(
    State(global): State<crate::Global>,
    Path((repo, arch, file_name)): Path<(String, String, String)>,
    req: Request<Body>,
) -> crate::Result<impl IntoResponse> {
    let repo_dir = global.config.data_dir.join("repos").join(&repo);

    let file_name =
        if file_name == format!("{}.db", repo) || file_name == format!("{}.db.tar.gz", repo) {
            format!("{}.db.tar.gz", arch)
        } else if file_name == format!("{}.files", repo)
            || file_name == format!("{}.files.tar.gz", repo)
        {
            format!("{}.files.tar.gz", arch)
        } else {
            file_name
        };

    Ok(ServeFile::new(repo_dir.join(file_name)).oneshot(req).await)
}

async fn post_package_archive(
    State(global): State<crate::Global>,
    Path(repo): Path<String>,
    body: Body,
) -> crate::Result<()> {
    let body = body.into_data_stream();
    let body = body.map_err(std::io::Error::other);
    let mut body = StreamReader::new(body);
    global
        .repo_manager
        .write()
        .await
        .add_pkg_from_reader(&global.db, &mut body, &repo)
        .await?;

    Ok(())

    //tracing::info!("Added '{}' to repository '{}'", pkg.file_name(), repo);
}

async fn delete_repo(
    State(global): State<crate::Global>,
    Path(repo): Path<String>,
) -> crate::Result<StatusCode> {
    let repo_removed = global
        .repo_manager
        .write()
        .await
        .remove_repo(&global.db, &repo)
        .await?;

    if repo_removed {
        tracing::info!("Removed repository '{}'", repo);

        Ok(StatusCode::OK)
    } else {
        Ok(StatusCode::NOT_FOUND)
    }
}

async fn delete_arch_repo(
    State(global): State<crate::Global>,
    Path((repo, arch)): Path<(String, String)>,
) -> crate::Result<StatusCode> {
    let repo_removed = global
        .repo_manager
        .write()
        .await
        .remove_repo_arch(&global.db, &repo, &arch)
        .await?;

    if repo_removed {
        tracing::info!("Removed arch '{}' from repository '{}'", arch, repo);

        Ok(StatusCode::OK)
    } else {
        Ok(StatusCode::NOT_FOUND)
    }
}

async fn delete_package(
    State(global): State<crate::Global>,
    Path((repo, arch, file_name)): Path<(String, String, String)>,
) -> crate::Result<StatusCode> {
    Ok(StatusCode::NOT_FOUND)
    //global.repo_manager.write().unwrap().remove_pkg(&global.db, &repo, &arch, name)
    //let clone = Arc::clone(&global.repo_manager);
    //let path = PathBuf::from(&repo).join(arch).join(&file_name);
    //
    //let res = tokio::task::spawn_blocking(move || {
    //    clone.write().unwrap().remove_pkg_from_path(path, true)
    //})
    //.await??;
    //
    //if let Some((name, version, release, arch)) = res {
    //    let res = db::query::repo::by_name(&global.db, &repo).await?;
    //
    //    if let Some(repo_entry) = res {
    //        let res =
    //            db::query::package::by_fields(&global.db, repo_entry.id, &arch, &name).await?;
    //
    //        if let Some(entry) = res {
    //            entry.delete(&global.db).await?;
    //        }
    //    }
    //
    //    tracing::info!("Removed '{}' from repository '{}'", file_name, repo);
    //
    //    Ok(StatusCode::OK)
    //} else {
    //    Ok(StatusCode::NOT_FOUND)
    //}
}
